"""
Setup script for obsplus
"""
import glob
import os
import stat
from collections import defaultdict
from os.path import join, exists, isdir
from pathlib import Path
import shutil

from setuptools import setup
from setuptools.command.develop import develop

here = Path(__file__).absolute().parent
version_file = here / "obsplus" / "version.py"

# --- get version
with version_file.open() as fi:
    content = fi.read().split("=")[-1].strip()
    __version__ = content.replace('"', "").replace("'", "")

# --- get readme
with open("README.rst") as readme_file:
    readme = readme_file.read()


# --- get sub-packages
def find_packages(base_dir="."):
    """ setuptools.find_packages wasn't working so I rolled this """
    out = []
    for fi in glob.iglob(join(base_dir, "**", "*"), recursive=True):
        if isdir(fi) and exists(join(fi, "__init__.py")):
            out.append(fi)
    out.append(base_dir)
    return out


def get_package_data_files():
    """ Gets data """
    data = Path("obsplus") / "datasets"
    out = defaultdict(list)
    # get list of datasets
    datasets = [x for x in data.glob("*") if x.is_dir()]
    for dataset in datasets:
        for ifile in dataset.glob("*"):
            if ifile.name.endswith("py") or ifile.name.endswith("pyc"):
                continue
            if ifile.is_dir():
                continue
            out[str(ifile.parent)].append(str(ifile))
    return list(out.items())


# --- requirements paths


def read_requirements(path):
    """ Read a requirements.txt file, return a list. """
    with Path(path).open("r") as fi:
        return fi.readlines()


package_req_path = here / "requirements.txt"
test_req_path = here / "tests" / "requirements.txt"
doc_req_path = here / "docs" / "requirements.txt"


class SetupDev(develop):
    """
    Install the obsplus git hook when development mode is run.
    These hooks run the black formatter on the code base and clear
    notebook outputs.
    """

    def run(self, *args, **kwargs):
        super().run(*args, **kwargs)
        # get path to get repo, do nothing if it doesnt exist
        gitpath = Path(".git")
        if not gitpath.exists():
            return
        # copy hooks
        hook_script = Path("scripts") / "pre-commit.py"
        assert hook_script.exists()
        # copy to git directory
        out_path = gitpath / "hooks" / "pre-commit"
        shutil.copy(hook_script, out_path)
        # make sure script is executable
        st = os.stat(str(out_path))
        os.chmod(out_path, st.st_mode | stat.S_IEXEC)


setup(
    name="obsplus",
    version=__version__,
    description="Some add-ons to obspy",
    long_description=readme,
    author="Derrick Chambers",
    author_email="djachambeador@gmail.com",
    url="https://bitbucket.org/smrd/obsplus",
    packages=find_packages("obsplus"),
    package_dir={"obsplus": "obsplus"},
    include_package_data=True,
    data_files=get_package_data_files(),
    license="GNU Lesser General Public License v3.0 or later (LGPLv3.0+)",
    zip_safe=False,
    keywords="obsplus",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
        "Programming Language :: Python :: 3.7",
    ],
    test_suite="tests",
    install_requires=read_requirements(package_req_path),
    tests_require=read_requirements(test_req_path),
    setup_requires=["pytest-runner>=2.0"],
    extras_require={"docs": read_requirements(doc_req_path)},
    cmdclass={"develop": SetupDev},
)
